﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Optimization {
  [Item("ResultParameter", "A parameter whose value is written to a result collection.")]
  [StorableClass]
  [NonDiscoverableType]
  public sealed class ResultParameter<T> : LookupParameter<ResultCollection>, IResultParameter<T> where T : class, IItem, new() {

    [Storable]
    private string resultName;
    public string ResultName {
      get { return resultName; }
      set {
        if (value == null) throw new ArgumentNullException();
        if (string.IsNullOrWhiteSpace(value)) {
          resultName = Name;
          OnResultNameChanged();
        } else if (!resultName.Equals(value)) {
          resultName = value;
          OnResultNameChanged();
        }
      }
    }

    [Storable]
    private T defaultValue;
    public T DefaultValue {
      get { return defaultValue; }
      set {
        if (value != defaultValue) {
          defaultValue = value;
          OnDefaultValueChanged();
        }
      }
    }

    public T ResultValue {
      get { return GetResultValue(); }
      set { SetResultValue(value); }
    }

    [StorableConstructor]
    private ResultParameter(bool deserializing) : base(deserializing) { }
    private ResultParameter(ResultParameter<T> original, Cloner cloner)
      : base(original, cloner) {
      resultName = original.resultName;
      defaultValue = cloner.Clone(original.defaultValue);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ResultParameter<T>(this, cloner);
    }
    public ResultParameter() : this("Anonymous") { }
    public ResultParameter(string name)
      : base(name, string.Empty, "Results") {
      resultName = Name;
      defaultValue = new T();
    }
    public ResultParameter(string name, string description) : this(name, description, new T()) { }
    public ResultParameter(string name, string description, T defaultValue)
      : base(name, description, "Results") {
      if (defaultValue == null) throw new ArgumentNullException("defaultValue");
      resultName = Name;
      this.defaultValue = defaultValue;
    }
    public ResultParameter(string name, string description, string resultName) : this(name, description, new T(), resultName) { }
    public ResultParameter(string name, string description, T defaultValue, string resultName)
      : base(name, description, "Results") {
      if (defaultValue == null) throw new ArgumentNullException("defaultValue");
      this.resultName = string.IsNullOrWhiteSpace(resultName) ? Name : resultName;
      this.defaultValue = defaultValue;
    }

    public override string ToString() {
      if (Name.Equals(ResultName)) return Name;
      return Name + ": " + ResultName;
    }

    private T GetResultValue() {
      var results = ActualValue;
      if (results == null)
        throw new InvalidOperationException("ResultParameter (" + ResultName + "): ResultCollection " + ActualName + " not found.");

      if (!results.ContainsKey(ResultName)) results.Add(new Result(ResultName, (T)DefaultValue.Clone()));
      var resultValue = results[ResultName].Value as T;
      if (resultValue == null)
        throw new InvalidOperationException(string.Format("Type mismatch. Result \"{0}\" does not contain a \"{1}\".", ResultName, typeof(T).GetPrettyName()));

      return resultValue;
    }

    private void SetResultValue(T value) {
      var results = ActualValue;
      if (results == null)
        throw new InvalidOperationException("ResultParameter (" + ResultName + "): ResultCollection " + ActualName + " not found.");

      if (!results.ContainsKey(ResultName))
        results.Add(new Result(ResultName, value));
      else results[ResultName].Value = value;
    }

    public event EventHandler ResultNameChanged;
    private void OnResultNameChanged() {
      var handler = ResultNameChanged;
      if (handler != null) handler(this, EventArgs.Empty);
      OnToStringChanged();
    }

    public event EventHandler DefaultValueChanged;
    private void OnDefaultValueChanged() {
      EventHandler handler = DefaultValueChanged;
      if (handler != null) handler(this, EventArgs.Empty);
      OnItemImageChanged();
    }
  }
}
